﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using MikuMikuDance.Model.Ver1;
using MikuMikuDance.XNA.Misc;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework;
using System.IO;

namespace MikuMikuDance.XNA.Model
{
    //MMDは面ごとにチャンネルが変化するため、XNAのメッシュビルダーでは力不足。
    //メタセコイアインポータを参考に作成する。
    static class MMDMeshBuilder
    {
        public static MeshContent BuildMesh(MMDModel1 model, string filename)
        {
            //メッシュ作成
            MeshContent buildingMesh = new MeshContent();
            //まずは頂点を登録。
            //ひげねこ氏によるとモデルごとのローカル座標である必要があるようだが
            //pmdはモデル一つ＝変換必要なし
            foreach (var vec in model.Vertexes)
                buildingMesh.Positions.Add(MMDXMath.ToVector3(vec.Pos));

            //ジオメトリとマテリアルの作成
            //メモ：頂点が3つ合わさって面、面が幾つか集まってジオメトリ。ジオメトリにマテリアルが付随。ジオメトリの集合がメッシュ
            long FaceIndex = 0;
            //ジオメトリとマテリアルの生成
            for (int i = 0; i < model.Materials.Length; i++)
            {
                GeometryContent geometry = new GeometryContent();
                BasicMaterialContent material = new BasicMaterialContent();
                geometry.Material = material;
                //マテリアル設定
                material.VertexColorEnabled = false;//頂点カラー無し
                material.Alpha = model.Materials[i].Alpha;
                material.DiffuseColor = MMDXMath.ToVector3(model.Materials[i].DiffuseColor);
                material.EmissiveColor = MMDXMath.ToVector3(model.Materials[i].MirrorColor);
                material.SpecularColor = MMDXMath.ToVector3(model.Materials[i].SpecularColor);
                material.SpecularPower = model.Materials[i].Specularity;
                if (!string.IsNullOrEmpty(model.Materials[i].TextureFileName))
                    material.Texture = new ExternalReference<TextureContent>(NormalizeFilepath(model.Materials[i].TextureFileName, filename));
                if (!string.IsNullOrEmpty(model.Materials[i].SphereTextureFileName))
                    material.Textures.Add("Sphere", new ExternalReference<TextureContent>(NormalizeFilepath(model.Materials[i].SphereTextureFileName, filename)));
                //一応その他マテリアル情報突っ込んでおく
                material.OpaqueData.Add("ToonIndex", model.Materials[i].ToonIndex);
                material.OpaqueData.Add("EdgeFlag", model.Materials[i].EdgeFlag);
                //ジオメトリのチャンネル設定
                //法線
                geometry.Vertices.Channels.Add(VertexChannelNames.Normal(0), typeof(Vector3), null);
                //テクスチャ
                if (!string.IsNullOrEmpty(model.Materials[i].TextureFileName))
                    geometry.Vertices.Channels.Add(VertexChannelNames.TextureCoordinate(0), typeof(Vector2), null);
                //ボーンウェイト
                geometry.Vertices.Channels.Add(VertexChannelNames.Weights(0), typeof(BoneWeightCollection), null);
                //面と頂点をジオメトリに登録
                //このマテリアルに対応する面は今までのマテリアルの面数の合計からこのマテリアルの面数個分
                for (long j = FaceIndex; j < FaceIndex + model.Materials[i].FaceVertCount; j++)
                {
                    //面に対応する頂点インデックスを取得(FaceVertexesは3要素ごとに1面に相当)
                    ushort posIdx = model.FaceVertexes[j];
                    //ジオメトリに登録
                    int currentIndex = geometry.Vertices.Add(posIdx);
                    int channelIdx=0;
                    //法線登録
                    geometry.Vertices.Channels.Get<Vector3>(channelIdx++)[currentIndex] = MMDXMath.ToVector3(model.Vertexes[posIdx].NormalVector);
                    //テクスチャ座標
                    if (!string.IsNullOrEmpty(model.Materials[i].TextureFileName))
                        geometry.Vertices.Channels.Get<Vector2>(channelIdx++)[currentIndex] = MMDXMath.ToVector2(model.Vertexes[posIdx].UV);
                    //ボーンウェイト
                    BoneWeightCollection boneWeight = new BoneWeightCollection();
                    int boneNum = model.Vertexes[posIdx].BoneNum[0];
                    if (boneNum >= 0 && boneNum < model.Bones.Length)
                        boneWeight.Add(new BoneWeight(model.Bones[boneNum].BoneName, model.Vertexes[posIdx].BoneWeight / 100f));
                    boneNum = model.Vertexes[posIdx].BoneNum[1];
                    if (boneNum >= 0 && boneNum < model.Bones.Length)
                        boneWeight.Add(new BoneWeight(model.Bones[boneNum].BoneName, 1.0f - model.Vertexes[posIdx].BoneWeight / 100f));
                    geometry.Vertices.Channels.Get<BoneWeightCollection>(channelIdx++)[currentIndex] = boneWeight;
                    geometry.Indices.Add(currentIndex);
                }
                //ジオメトリをモデルに追加
                buildingMesh.Geometry.Add(geometry);
                //面頂点カウント進める
                FaceIndex += model.Materials[i].FaceVertCount;
            }
            //重複頂点データのマージ
            MeshHelper.MergeDuplicateVertices(buildingMesh);
            //メッシュ出来たので返却
            return buildingMesh;
        }

        private static string NormalizeFilepath(string filename, string modelfilename)
        {
            if (Path.IsPathRooted(filename))
                return filename;
            return Path.Combine(Path.GetDirectoryName(modelfilename), filename);
        }
    }
}
