﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace MikuMikuDance.XNA.Misc
{

    /// <summary>
    /// クォータニオンを使った回転と平行移動を表すことのできる頂点変換用構造体
    /// 通常の行列の半分以下のメモリ使用量になる
    /// </summary>
    public struct QuatTransform
    {
        #region フィールド

        /// <summary>
        /// 回転
        /// </summary>
        public Quaternion Rotation;

        /// <summary>
        /// 平行移動
        /// </summary>
        public Vector3 Translation;

        #endregion
        /// <summary>
        /// 恒等QuatTransformを返します
        /// </summary>
        public static QuatTransform Identity { get { return new QuatTransform(Quaternion.Identity, Vector3.Zero); } }
        #region 初期化

        /// <summary>
        /// クォータニオンと平行移動を指定して生成する
        /// </summary>
        /// <param name="rotation"></param>
        /// <param name="translation"></param>
        public QuatTransform(Quaternion rotation, Vector3 translation)
        {
            Rotation = rotation;
            Translation = translation;
        }

        /// <summary>
        /// 指定された行列から生成する
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static QuatTransform FromMatrix(Matrix matrix)
        {
            // 行列の分解
            Quaternion rotation;
            Vector3 translation;
            Vector3 scale;
            matrix.Decompose(out scale, out rotation, out translation);

            // 一意のスケールか？
            if (!CloseEnough(scale.X, scale.Y) || !CloseEnough(scale.X, scale.Z))
            {
                throw new InvalidOperationException(
                    "一意のスケール(X,Y,Zが同じスケール値)ではありません");
            }

            if (!CloseEnough(scale.X, 1.0f))
                throw new InvalidOperationException("スケール値が1以外です");

            return new QuatTransform(rotation, translation);
        }

        static bool CloseEnough(float a, float b)
        {
            return (System.Math.Abs(a - b) < 1e-4f);
        }

        #endregion

        /// <summary>
        /// QuatTransformの結合
        /// </summary>
        public static void Multiply(ref QuatTransform value1, ref QuatTransform value2, out QuatTransform result)
        {
            // 平行移動の算出
            Vector3 newTranslation;
            value2.Rotation.Normalize();
            Vector3.Transform(ref value1.Translation, ref value2.Rotation,
                                out newTranslation);

            newTranslation.X += value2.Translation.X;
            newTranslation.Y += value2.Translation.Y;
            newTranslation.Z += value2.Translation.Z;

            // 回転部分の結合
            Quaternion.Concatenate(ref value1.Rotation, ref value2.Rotation,
                                        out result.Rotation);

            result.Translation = newTranslation;
            //正規化
            result.Rotation.Normalize();

        }
        /// <summary>
        /// QuatTransformの結合
        /// </summary>
        public static QuatTransform operator *(QuatTransform value1, QuatTransform value2)
        {
            QuatTransform result;
            QuatTransform.Multiply(ref value1, ref value2, out result);
            return result;
        }
        internal Matrix CreateMatrix()
        {
            Matrix move = Matrix.CreateTranslation(Translation);
            Matrix rot = Matrix.CreateFromQuaternion(Rotation);
            Matrix result;
            Matrix.Multiply(ref rot, ref move, out result);
            return result;//回転・移動の順にかけ合わせる
        }
        public void CreateMatrix(out Matrix result)
        {
            float ww = Rotation.W * Rotation.W - 0.5f;
            float num11 = ww + Rotation.X * Rotation.X;
            float num12 = Rotation.X * Rotation.Y + Rotation.W * Rotation.Z;
            float num13 = Rotation.X * Rotation.Z - Rotation.W * Rotation.Y;

            float num21 = Rotation.X * Rotation.Y + Rotation.W * Rotation.Z;
            float num22 = ww - Rotation.Y * Rotation.Y;
            float num23 = Rotation.Y * Rotation.Z + Rotation.W * Rotation.X;

            float num31 = Rotation.X * Rotation.Z + Rotation.W * Rotation.Y;
            float num32 = Rotation.Y * Rotation.Z - Rotation.W * Rotation.X;
            float num33 = ww + Rotation.Z * Rotation.Z;

            result.M11 = num11 * 2f;
            result.M12 = num12 * 2f;
            result.M13 = num13 * 2f;
            result.M14 = 0;
            result.M21 = num21 * 2f;
            result.M22 = num22 * 2f;
            result.M23 = num23 * 2f;
            result.M24 = 0;
            result.M31 = num31 * 2f;
            result.M32 = num32 * 2f;
            result.M33 = num33 * 2f;
            result.M34 = 0;
            result.M41 = Translation.X;
            result.M42 = Translation.Y;
            result.M43 = Translation.Z;
            result.M44 = 1;
        }
        /// <summary>
        /// 移動QuatTransformを返します
        /// </summary>
        /// <param name="x">X移動</param>
        /// <param name="y">Y移動</param>
        /// <param name="z">Z移動</param>
        /// <returns>移動QuatTransform</returns>
        public static QuatTransform CreateTranslation(float x, float y, float z)
        {
            return new QuatTransform(Quaternion.Identity, new Vector3(x, y, z));
        }
        /// <summary>
        /// 移動QuatTransformを返します
        /// </summary>
        /// <param name="x">X移動</param>
        /// <param name="y">Y移動</param>
        /// <param name="z">Z移動</param>
        /// <param name="result">移動QuatTransform</param>
        public static void CreateTranslation(float x, float y, float z, out QuatTransform result)
        {
            result = new QuatTransform(Quaternion.Identity, new Vector3(x, y, z));
        }

        /// <summary>
        /// X軸回転QuatTransformを返します
        /// </summary>
        /// <param name="rot">X回転</param>
        /// <returns>X軸回転QuatTransform</returns>
        public static QuatTransform CreateRotationX(float rot)
        {
            return new QuatTransform(Quaternion.CreateFromYawPitchRoll(0, rot, 0), Vector3.Zero);
        }
        /// <summary>
        /// X軸回転QuatTransformを返します
        /// </summary>
        /// <param name="rot">X回転</param>
        /// <param name="result">X軸回転QuatTransform</param>
        public static void CreateRotationX(float rot, out QuatTransform result)
        {
            result = new QuatTransform();
            Quaternion.CreateFromYawPitchRoll(0, rot, 0, out result.Rotation);
        }
        /// <summary>
        /// Y軸回転QuatTransformを返します
        /// </summary>
        /// <param name="rot">Y回転</param>
        /// <returns>Y軸回転QuatTransform</returns>
        public static QuatTransform CreateRotationY(float rot)
        {
            return new QuatTransform(Quaternion.CreateFromYawPitchRoll(rot, 0, 0), Vector3.Zero);
        }
        /// <summary>
        /// Y軸回転QuatTransformを返します
        /// </summary>
        /// <param name="rot">Y回転</param>
        /// <param name="result">Y軸回転QuatTransform</param>
        public static void CreateRotationY(float rot, out QuatTransform result)
        {
            result = new QuatTransform();
            Quaternion.CreateFromYawPitchRoll(rot, 0, 0, out result.Rotation);
        }
        /// <summary>
        /// Z軸回転QuatTransformを返します
        /// </summary>
        /// <param name="rot">Z回転</param>
        /// <returns>Z軸回転QuatTransform</returns>
        public static QuatTransform CreateRotationZ(float rot)
        {
            return new QuatTransform(Quaternion.CreateFromYawPitchRoll(0, 0, rot), Vector3.Zero);
        }
        /// <summary>
        /// Z軸回転QuatTransformを返します
        /// </summary>
        /// <param name="rot">Z回転</param>
        /// <param name="result">Z軸回転QuatTransform</param>
        public static void CreateRotationZ(float rot, out QuatTransform result)
        {
            result = new QuatTransform();
            Quaternion.CreateFromYawPitchRoll(0, 0, rot, out result.Rotation);
        }

        /// <summary>
        /// 姿勢の線形補間
        /// </summary>
        /// <param name="pose1">姿勢1</param>
        /// <param name="pose2">姿勢2</param>
        /// <param name="amount">補完係数(0-1)</param>
        /// <param name="result">線形補間された姿勢</param>
        /// <remarks>Quaternionは球状線形補間を使用</remarks>
        internal static void Lerp(ref QuatTransform pose1, ref QuatTransform pose2, float amount, out QuatTransform result)
        {
            Vector3.Lerp(ref pose1.Translation, ref pose2.Translation, amount, out result.Translation);
            Quaternion.Slerp(ref pose1.Rotation, ref pose2.Rotation, amount, out result.Rotation);
        }
    }
}
