﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;

namespace MikuMikuDance.XNA.Model
{
    /// <summary>
    /// モデルパーツ
    /// </summary>
    public abstract class MMDModelPart
    {
        /// <summary>
        /// モデルのトライアングル数
        /// </summary>
        protected readonly int triangleCount;
        /// <summary>
        /// モデルの頂点数
        /// </summary>
        protected readonly int vertexCount;
        /// <summary>
        /// インデックスバッファ
        /// </summary>
        protected IndexBuffer indexBuffer;
        /// <summary>
        /// 頂点バッファ
        /// </summary>
        protected DynamicVertexBuffer vertexBuffer;//このフィールドは継承先で代入
        /// <summary>
        /// エフェクト
        /// </summary>
        public BasicEffect Effect { get; internal set; }

        static BlendState ModelBlendState = BlendState.AlphaBlend;
        static void SetUpRenderState(GraphicsDevice GraphicsDevice)
        {
            GraphicsDevice.BlendState = ModelBlendState;
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.SamplerStates[0] = SamplerState.LinearWrap;
        }
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertexCount">頂点数</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPart(int triangleCount, int vertexCount, IndexBuffer indexBuffer)
        {
            this.triangleCount = triangleCount;
            this.vertexCount = vertexCount;
            this.indexBuffer = indexBuffer;

        }
        /// <summary>
        /// エフェクトにマトリックスを適用
        /// </summary>
        /// <param name="world">ワールド</param>
        /// <param name="view">ビュー</param>
        /// <param name="projection">プロジェクション</param>
        public void SetMatrices(ref Matrix world, ref Matrix view, ref Matrix projection)
        {
            IEffectMatrices matrices = Effect as IEffectMatrices;
            if (matrices != null)
            {
                matrices.World = world;
                matrices.View = view;
                matrices.Projection = projection;
            }
        }
        /// <summary>
        /// モデルの描画
        /// </summary>
        public void Draw()
        {
            GraphicsDevice graphics = Effect.GraphicsDevice;
            //レンダーステートセットアップ
            SetUpRenderState(graphics);
            //バッファセット
            graphics.Indices = indexBuffer;
            graphics.SetVertexBuffer(vertexBuffer);

            foreach (EffectPass pass in Effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                graphics.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, vertexCount, 0, triangleCount);
            }

            graphics.Indices = null;
            graphics.SetVertexBuffer(null);
        }

        //ファクトリー関数
        internal static MMDModelPart Create(int triangleCount, MMDVertex[] Vertices, IndexBuffer indexBuffer)
        {
            if (Vertices is MMDVertexNm[])
            {
                if (Vertices is MMDVertexNmTx[])
                {
                    if (Vertices is MMDVertexNmTxVc[])
                        return new MMDModelPartPNmTxVc(triangleCount, (MMDVertexNmTxVc[])Vertices, indexBuffer);
                    else
                        return new MMDModelPartPNmTx(triangleCount, (MMDVertexNmTx[])Vertices, indexBuffer);
                }
                else
                {
                    if (Vertices is MMDVertexNmVc[])
                        return new MMDModelPartPNmVc(triangleCount, (MMDVertexNmVc[])Vertices, indexBuffer);
                    else
                        return new MMDModelPartPNm(triangleCount, (MMDVertexNm[])Vertices, indexBuffer);
                }
            }
            else
            {
                if (Vertices is MMDVertexTx[])
                {
                    if (Vertices is MMDVertexTxVc[])
                        return new MMDModelPartPTxVc(triangleCount, (MMDVertexTxVc[])Vertices, indexBuffer);
                    else
                        return new MMDModelPartPTx(triangleCount, (MMDVertexTx[])Vertices, indexBuffer);
                }
                else
                {
                    if (Vertices is MMDVertexVc[])
                        return new MMDModelPartPVc(triangleCount, (MMDVertexVc[])Vertices, indexBuffer);
                    else
                        return new MMDModelPartP(triangleCount, Vertices, indexBuffer);
                }
            }
        }

        internal abstract void SetBone(Matrix[] bones);

        
    }
    /// <summary>
    /// モデルパーツ(頂点のみ)
    /// </summary>
    public class MMDModelPartP : MMDModelPart
    {
        readonly VertexPosition[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertex[] vertices;
        
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartP(int triangleCount, MMDVertex[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPosition[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPosition), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
    /// <summary>
    /// モデルパーツ(頂点+色)
    /// </summary>
    public class MMDModelPartPVc : MMDModelPart
    {
        readonly VertexPositionColor[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexVc[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPVc(int triangleCount, MMDVertexVc[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionColor[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionColor), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].Color = new Color(vertices[i].VertexColor);
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position));
            //頂点位置の計算
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
    /// <summary>
    /// モデルパーツ(頂点+法線)
    /// </summary>
    public class MMDModelPartPNm : MMDModelPart
    {//スフィアマップだけ貼りつけている物はこれになる……
        readonly VertexPositionNormal[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexNm[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPNm(int triangleCount, MMDVertexNm[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionNormal[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionNormal), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].Normal = vertices[i].Normal;
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点と法線位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
    /// <summary>
    /// モデルパーツ(頂点+テクスチャー)
    /// </summary>
    public class MMDModelPartPTx : MMDModelPart
    {
        readonly VertexPositionTexture[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexTx[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPTx(int triangleCount, MMDVertexTx[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionTexture[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionTexture), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].TextureCoordinate = vertices[i].TextureCoordinate;
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }

    /// <summary>
    /// モデルパーツ(頂点+法線+色)
    /// </summary>
    public class MMDModelPartPNmVc : MMDModelPart
    {
        readonly VertexPositionNormalColor[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexNmVc[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPNmVc(int triangleCount, MMDVertexNmVc[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionNormalColor[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionNormalColor), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].Normal = vertices[i].Normal; 
                gpuVertices[i].Color = new Color(vertices[i].VertexColor);
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点と法線位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
    /// <summary>
    /// モデルパーツ(頂点+テクスチャー+カラー)
    /// </summary>
    public class MMDModelPartPTxVc : MMDModelPart
    {
        readonly VertexPositionColorTexture[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexTxVc[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPTxVc(int triangleCount, MMDVertexTxVc[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionColorTexture[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionColorTexture), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].TextureCoordinate = vertices[i].TextureCoordinate;
                gpuVertices[i].Color = new Color(vertices[i].VertexColor);
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
    /// <summary>
    /// モデルパーツ(頂点+法線+テクスチャー)
    /// </summary>
    public class MMDModelPartPNmTx : MMDModelPart
    {
        readonly VertexPositionNormalTexture[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexNmTx[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPNmTx(int triangleCount, MMDVertexNmTx[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionNormalTexture[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionNormalTexture), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].Normal = vertices[i].Normal; 
                gpuVertices[i].TextureCoordinate = vertices[i].TextureCoordinate;
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点と法線位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
    /// <summary>
    /// モデルパーツ(頂点+法線+テクスチャー+カラー)
    /// </summary>
    public class MMDModelPartPNmTxVc : MMDModelPart
    {
        readonly VertexPositionNormalTextureColor[] gpuVertices;
        /// <summary>
        /// 頂点配列
        /// </summary>
        protected MMDVertexNmTxVc[] vertices;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="triangleCount">三角形の個数</param>
        /// <param name="vertices">頂点配列</param>
        /// <param name="indexBuffer">インデックスバッファ</param>
        public MMDModelPartPNmTxVc(int triangleCount, MMDVertexNmTxVc[] vertices, IndexBuffer indexBuffer)
            : base(triangleCount, vertices.Length, indexBuffer)
        {
            this.vertices = vertices;
            //GPUリソース作成
            gpuVertices = new VertexPositionNormalTextureColor[vertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionNormalTextureColor), vertices.Length, BufferUsage.WriteOnly);
            //初期値代入
            for (int i = 0; i < vertices.Length; i++)
            {
                gpuVertices[i].Position = vertices[i].Position;
                gpuVertices[i].Normal = vertices[i].Normal;
                gpuVertices[i].Color = new Color(vertices[i].VertexColor); 
                gpuVertices[i].TextureCoordinate = vertices[i].TextureCoordinate;
            }
            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
        internal override void SetBone(Matrix[] bones)
        {
            //頂点と法線位置の計算
            System.Threading.Tasks.Parallel.For(0, vertices.Length,
                (i) => SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal));
            /*for (int i = 0; i < vertices.Length; i++)
            {
                SkinningHelpers.SkinVertex(bones, vertices[i], out gpuVertices[i].Position, out gpuVertices[i].Normal);
            }*/
            //GPUバッファへの書き込み
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }
    }
}
