﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MikuMikuDance.XNA.Misc;
using Microsoft.Xna.Framework;
using System.Diagnostics;

namespace MikuMikuDance.XNA.Motion
{
    /// <summary>
    /// モーショントラック。モーションの再生管理及びポーズ差分の計算を行う
    /// </summary>
    public class MMDMotionTrack
    {
        bool bLoopPlay = false;
        bool bReverse = false;
        long LastUpdate = 0;
        decimal NowFrame = 0;
        decimal MaxFrame = 0;
        Stopwatch stopwatch = new Stopwatch();
        //モーションデータ
        Dictionary<string, List<MMDBoneKeyFrame>> boneFrames;
        Dictionary<string, List<MMDFaceKeyFrame>> faceFrames;
        //モーションデータの読み出し位置
        Dictionary<string, int> bonePos = new Dictionary<string, int>();
        Dictionary<string, int> facePos = new Dictionary<string, int>();
        //ブレンディングファクター
        float m_blendingFactor = 1;

        //トラックから抽出されたボーンの差分一覧
        internal Dictionary<string, SQTTransform> SubPoses { get; private set; }

        /// <summary>
        /// ブレンディングファクター
        /// </summary>
        /// <remarks>このトラックのモーションをどの程度モデルに適応するか。0～1の範囲の値</remarks>
        public float BoneBlendingFactor { 
            get { return m_blendingFactor; }
            set { m_blendingFactor = MathHelper.Clamp(value, 0, 1); }
        }
        /// <summary>
        /// モーション再生用FPMs
        /// </summary>
        public decimal FramePerMillisecond { get; set; }
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="motionData">モーションデータ</param>
        public MMDMotionTrack(MMDMotionData motionData)
        {
            //ボーンの配列抜き出し
            boneFrames = motionData.BoneFrames;
            //表情の配列抜き出し
            faceFrames = motionData.FaceFrames;
            //モーションのFPS=30
            FramePerMillisecond = 30m / 1000m;
            //差分一覧を作成
            SubPoses = new Dictionary<string, SQTTransform>(motionData.BoneFrames.Count);
            //現在の再生位置を設定&最大フレーム数のチェック
            foreach (var it in motionData.BoneFrames)
            {
                bonePos.Add(it.Key, 0);
                foreach (var it2 in it.Value)
                {
                    if (it2.FrameNo > MaxFrame)
                        MaxFrame = it2.FrameNo;
                }
            }
            foreach (var it in motionData.FaceFrames)
            {
                facePos.Add(it.Key, 0);
                foreach (var it2 in it.Value)
                {
                    if (it2.FrameNo > MaxFrame)
                        MaxFrame = it2.FrameNo;
                }
            }
        }

        /// <summary>
        /// モーションの再生
        /// </summary>
        /// <param name="LoopPlay"></param>
        public void Start(bool LoopPlay = false)
        {
            bLoopPlay = LoopPlay;
            stopwatch.Start();
            LastUpdate = stopwatch.ElapsedTicks;
        }
        /// <summary>
        /// モーションの停止
        /// </summary>
        public void Stop()
        {
            InnerStop();
            TimeUpdate();
        }
        void InnerStop()
        {
            stopwatch.Stop();
        }
        /// <summary>
        /// 巻き戻し及び逆再生設定
        /// </summary>
        /// <param name="Reverse"></param>
        public void Reset(bool Reverse = false)
        {
            bReverse = Reverse;
            stopwatch.Stop();
            stopwatch.Reset();
            LastUpdate = 0;
            Rewind();
        }
        //更新
        internal void Update()
        {
            TimeUpdate();
            SubPoses.Clear();
            //ボーンの更新
            foreach (var frameList in boneFrames)
            {
                //カーソル位置の更新
                int CursorPos = bonePos[frameList.Key];
                if (!bReverse)
                    for (; CursorPos < frameList.Value.Count && frameList.Value[CursorPos].FrameNo < NowFrame; ++CursorPos) ;
                else
                    for (; CursorPos > 0 && frameList.Value[CursorPos - 1].FrameNo > NowFrame; --CursorPos) ;
                bonePos[frameList.Key] = CursorPos;
                if (CursorPos == 0 || CursorPos == frameList.Value.Count)
                    continue;//このボーンの再生終わり
                //時間経過取得
                float Progress = ((float)NowFrame - (float)frameList.Value[CursorPos - 1].FrameNo) / ((float)frameList.Value[CursorPos].FrameNo - (float)frameList.Value[CursorPos - 1].FrameNo);
                SQTTransform subPose;
                MMDBoneKeyFrame pose1 = frameList.Value[CursorPos - 1], pose2 = frameList.Value[CursorPos];
                MMDBoneKeyFrame.Lerp(pose1, pose2, Progress, out subPose);
                SubPoses.Add(frameList.Key, subPose);
            }
        }

        private void TimeUpdate()
        {
            decimal ElapsedTime;
            ElapsedTime = (decimal)(stopwatch.ElapsedTicks - LastUpdate) * 1000m / (decimal)Stopwatch.Frequency;
            if (!bReverse)
            {
                NowFrame += ElapsedTime * FramePerMillisecond;
                if (NowFrame > MaxFrame)
                {
                    if (bLoopPlay)
                        Rewind();
                    else
                        InnerStop();
                }
            }
            else
            {
                NowFrame -= ElapsedTime * FramePerMillisecond;
                if (NowFrame < 0)
                {
                    if (bLoopPlay)
                        Rewind();
                    else
                        InnerStop();
                }
            }
            LastUpdate = stopwatch.ElapsedTicks;

        }
        //巻き戻し処理
        private void Rewind()
        {
            if (!bReverse)
            {
                while (NowFrame >= MaxFrame)
                    NowFrame -= MaxFrame;
                //ボーンの更新
                foreach (var frameList in boneFrames)
                    bonePos[frameList.Key] = 0;
                foreach (var faceList in faceFrames)
                    facePos[faceList.Key] = 0;
            }
            else
            {
                while (NowFrame <= 0)
                    NowFrame += MaxFrame;
                //ボーンの更新
                foreach (var frameList in boneFrames)
                    bonePos[frameList.Key] = frameList.Value.Count - 1;
                foreach (var faceList in faceFrames)
                    facePos[faceList.Key] = faceList.Value.Count - 1;
            }
        }
    }
}
