using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using MikuMikuDance.XNA.Model;
using MikuMikuDance.XNA;
using MikuMikuDance.XNA.Motion;
using DebugSample;
using MikuMikuDance.XNA.Misc;

namespace MikuMikuDanceXNADemo1
{
    /// <summary>
    /// 基底 Game クラスから派生した、ゲームのメイン クラスです。
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        MMDModel model;
        MMDMotionData motion;

        bool flag = false;

        // デバッグマネージャー
        DebugManager debugManager;

        // デバッグコマンドUI
        DebugCommandUI debugCommandUI;

        // FPSカウンター
        FpsCounter fpsCounter;

        // タイムルーラー
        TimeRuler timerRuler;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
            IsFixedTimeStep = false;
        }

        /// <summary>
        /// ゲームが実行を開始する前に必要な初期化を行います。
        /// ここで、必要なサービスを照会して、関連するグラフィック以外のコンテンツを
        /// 読み込むことができます。base.Initialize を呼び出すと、使用するすべての
        /// コンポーネントが列挙されるとともに、初期化されます。
        /// </summary>
        protected override void Initialize()
        {
            // デバッグマネージャーの初期化と追加
            debugManager = new DebugManager(this);
            Components.Add(debugManager);

            // デバッグマコマンドUIの初期化と追加
            debugCommandUI = new DebugCommandUI(this);

            // デバッグコマンドUIを最上面に表示させる為にDrawOrderを変更する
            debugCommandUI.DrawOrder = 100;

            Components.Add(debugCommandUI);

            // FPSカウンターの初期化と追加
            fpsCounter = new FpsCounter(this);
            Components.Add(fpsCounter);
            fpsCounter.Visible = true;
            
            // タイムルーラーの初期化と追加
            timerRuler = new TimeRuler(this);
            Components.Add(timerRuler);
            timerRuler.Visible = true;
            timerRuler.ShowLog = true;
            
#if WINDOWS || XBOX
            // リモートデバッグコマンド「remote」の追加
            Components.Add(new RemoteDebugCommand(this));
#endif
            base.Initialize();
        }

        /// <summary>
        /// LoadContent はゲームごとに 1 回呼び出され、ここですべてのコンテンツを
        /// 読み込みます。
        /// </summary>
        protected override void LoadContent()
        {
            // 新規の SpriteBatch を作成します。これはテクスチャーの描画に使用できます。
            spriteBatch = new SpriteBatch(GraphicsDevice);

            model = MMDXCore.Instance.LoadModel("Miku", Content);
            motion = MMDXCore.Instance.LoadMotion("TrueMyHeart", Content);

            model.AnimationPlayer.AddMotion("TrueMyHeart", motion);
            // TODO: this.Content クラスを使用して、ゲームのコンテンツを読み込みます。
            timerRuler.Visible = true;
            timerRuler.ShowLog = true;
            fpsCounter.Visible = true;
            MMDXProfiler.MMDBeginMark += new BeginMarkDelegate(BeginMark);
            MMDXProfiler.MMDEndMark += new EndMarkDelegate(EndMark);
        }
        void BeginMark(int index, string name, Color color) { timerRuler.BeginMark(index, name, color); }
        void EndMark(int index, string name) { timerRuler.EndMark(index, name); }

        /// <summary>
        /// UnloadContent はゲームごとに 1 回呼び出され、ここですべてのコンテンツを
        /// アンロードします。
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: ここで ContentManager 以外のすべてのコンテンツをアンロードします。
        }

        /// <summary>
        /// ワールドの更新、衝突判定、入力値の取得、オーディオの再生などの
        /// ゲーム ロジックを、実行します。
        /// </summary>
        /// <param name="gameTime">ゲームの瞬間的なタイミング情報</param>
        protected override void Update(GameTime gameTime)
        {
            // タイムルーラーにフレーム開始を伝える
            timerRuler.StartFrame();

            // 更新期間、"Update"の測定開始
            timerRuler.BeginMark("Update", Color.Blue);
            
            // ゲームの終了条件をチェックします。
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();
            if (!flag && Keyboard.GetState().IsKeyDown(Keys.Enter))
            {
                model.AnimationPlayer["TrueMyHeart"].Start();
                flag = true;
            }
            // TODO: ここにゲームのアップデート ロジックを追加します。
            model.Update();

            base.Update(gameTime);
            // 更新期間、"Update"の測定終了
            timerRuler.EndMark("Update");
        }

        /// <summary>
        /// ゲームが自身を描画するためのメソッドです。
        /// </summary>
        /// <param name="gameTime">ゲームの瞬間的なタイミング情報</param>
        protected override void Draw(GameTime gameTime)
        {
            // 描画期間、"Draw"の測定開始
            timerRuler.BeginMark("Draw", Color.Yellow);
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // TODO: ここに描画コードを追加します。
            /*foreach (var modelMesh in model.Meshes)
            {
                foreach (var part in modelMesh.MeshParts)
                {
                    BasicEffect effect = part.Effect as BasicEffect;
                    effect.EnableDefaultLighting();
                    effect.World = Matrix.Identity;
                    effect.View = Matrix.CreateLookAt(new Vector3(0, 10, 30), new Vector3(0, 10, 0), Vector3.UnitY);
                    Viewport viewport = GraphicsDevice.Viewport;
                    float aspectRatio = (float)viewport.Width / (float)viewport.Height;
                    effect.Projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, aspectRatio, 1, 300);
                }
                modelMesh.Draw();
            }*/
            Viewport viewport = GraphicsDevice.Viewport;
            float aspectRatio = (float)viewport.Width / (float)viewport.Height;
            model.Transform = Matrix.Identity;
            model.Draw(Matrix.CreateLookAt(new Vector3(0, 10, 70), new Vector3(0, 10, 0), Vector3.UnitY), Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, aspectRatio, 1, 300));
            base.Draw(gameTime);
            // 描画期間、"Draw"の測定終了
            timerRuler.EndMark("Draw");
        }
    }
}
